/*		fft.c		*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <math.h>
#include "sslib.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define N 4

double dt = 0.1;
double alpha = 1.5;
double beta = 2.0;
double gamma = 2.0/M_PI;
double hini = 1.5;
double v = 0.01;

void print(double *ar, double *ai, int n);
void print2(double *ar, double *ai, int n, int nmax);
void fft1(double ar[], double ai[], int n, int iter, int flag);
void fft2(double ar[], double ai[], int n, int nmax, int flag);
void array_insert(double a[], double b[]);
void do_fft2(double ar[], double ai[], double br[], double bi[], int flag);
void function(double br[], double bi[], double fr[], double fi[], double t);
void runge2(double br[], double bi[], double t);

int main(void)
{
	static double ar2[N*N] = { 0., 0., 0., 0., 0., 1., 1., 0.,
							  0., 1., 1., 0., 0., 0., 0., 0.};
	static double ai2[N*N] = { 0., 0., 0., 0., 0., 0., 0., 0.,
							  0., 0., 0., 0., 0., 0., 0., 0.};
	static double br2[N*N] = { 0., 0., 0., 0., 0., 0., 0., 0.,
							  0., 0., 0., 0., 0., 0., 0., 0.};
	static double bi2[N*N] = { 0., 0., 0., 0., 0., 0., 0., 0.,
							  0., 0., 0., 0., 0., 0., 0., 0.};
	int i;

	print2(ar2, ai2, N, N);

	for(i=0; i<40; ++i){
		do_fft2(ar2, ai2, br2, bi2, 0);
		runge2(br2, bi2, i*dt);
		do_fft2(br2, bi2, ar2, ai2, 1);
		print2(ar2, ai2, N, N);
	}
	return 1;
}


void fft1(double ar[], double ai[], int n, int iter, int flag)
{
	int i, it, j, j1, j2, k, xp, xp2;
	double arg, dr1, dr2, di1, di2, tr, ti, w, wr, wi;

	if(n < 2)
	{
		fprintf(stderr, "Error : n < 2  in fft1()\n");
		return;
	}
	if(iter <= 0)
	{
		iter = 0;
		i = n;
		while((i /= 2) != 0)	iter++;
	}
	j = 1;
	for(i = 0; i < iter; i++)	j *= 2;
	if(n != j)
	{
		fprintf(stderr, "Error : n != 2 ^ k  in fft1()\n");
		return;
	}
	w = (flag? M_PI: -M_PI) / (double)n;
	xp2 = n;
	for(it = 0; it < iter; it++)
	{
		xp = xp2;
		xp2 /= 2;
		w *= 2;
		for(k = 0, i = - xp; k < xp2; i++)
		{
			wr = cos(arg = w * k++);
			wi = sin(arg);
			for(j = xp; j <= n; j += xp)
			{
				j2 = (j1 = j + i) + xp2;
				tr = (dr1 = ar[j1]) - (dr2 = ar[j2]);
				ti = (di1 = ai[j1]) - (di2 = ai[j2]);
				ar[j1] = dr1 + dr2;
				ai[j1] = di1 + di2;
				ar[j2] = tr * wr - ti * wi;
				ai[j2] = ti * wr + tr * wi;
			}
		}
	}
	j = j1 = n / 2;
	j2 = n - 1;
	for(i = 1; i < j2; i++)
	{
		if(i < j)
		{
			w = ar[i];
			ar[i] = ar[j];
			ar[j] = w;
			w = ai[i];
			ai[i] = ai[j];
			ai[j] = w; 
		}
		k = j1;
		while(k <= j)
		{
			j -= k;
			k /= 2;
		}
		j += k;
	}
	if(flag == 0)	return;
	w = 1. / (double)n;
	for(i = 0; i < n; i++)
	{
		ar[i] *= w;
		ai[i] *= w;
	}
	return;
}



void fft2(double ar[], double ai[], int n, int nmax, int flag)
{
	int i, iter, j, k;
	double *p, *q, *wr, *wi;

	if(n < 2)
	{
		fprintf(stderr, "Error : Illegal parameter in fft2()\n");
		return;
	}
	iter = 0;
	i = n;
	while((i /= 2) != 0)	iter++;
	j = 1;
	for(i = 1; i <= iter; i++)	j *= 2;
	if(n != j)
	{
		fprintf(stderr, "Error : n != 2 ^ k  in fft2()\n");
		return;
	}
	wr = (double *)malloc(n * sizeof(double));
	if(wr == NULL)
	{
		fprintf(stderr, "Error : Out of memory  in fft2()\n");
		return;
	}
	wi = (double *)malloc(n * sizeof(double));
	if(wi == NULL)
	{
		fprintf(stderr, "Error : Out of memory  in fft2()\n");
		free((char *)wr);
		return;
	}

	for(j = k = 0; j < n; j++, k += nmax)
	{
		for(i = 0, p = ar + j, q = ai + j; i < n; i++, p += nmax, q += nmax)
		{
			*(wr + i) = *p;
			*(wi + i) = *q;
		}
		fft1(wr, wi, n, iter, flag);
		for(i = 0, p = ar + j, q = ai + j; i < n; i++, p += nmax, q += nmax)
		{
			*p = *(wr + i);
			*q = *(wi + i);
		}
	}
	for(i = k = 0; i < n; i++, k += nmax)
	{
		for(j = 0, p = ar + k, q = ai + k; j < n; j++)
		{
			*(wr + j) = *p++;
			*(wi + j) = *q++;
		}
		fft1(wr, wi, n, iter, flag);
		for(j = 0, p = ar + k, q = ai + k; j < n; j++)
		{
			*p++ = *(wr + j);
			*q++ = *(wi + j);
		}
	}
	free((char *)wr);
	free((char *)wi);
	return;
}



void print(double *ar, double *ai, int n)
{
	int i;
	double *p, *q;

	printf("    ar              ai\n");
	for(i = 0, p = ar, q = ai; i < n; i++)	printf("%14.7e %14.7e\n", *p++, *q++);
}

void print2(double *ar, double *ai, int n, int nmax)
{
	int i, j, k, l;
	double *p;

	printf("< ar >\n");
	for(i = k = 0, p = ar; i < n; i++)
	{
		for(j = 0; j < nmax; j++, k++)	printf("%14.7e ", *p++);
		putchar('\n');
	}
	printf("< ai >\n");
	for(i = k = 0, p = ai; i < n; i++)
	{
		for(j = 0; j < nmax; j++, k++)	printf("%14.7e ", *p++);
		putchar('\n');
	}
}


//܂ł4̊֐͏ÊłB


void array_insert(double a[], double b[]){//z̑s֐BabɑB
	int i;
	for(i=0; i<N*N; ++i){
		b[i]=a[i];
	}
}

void do_fft2(double ar[], double ai[], double br[], double bi[], int flag){//2fftsBãt[GϊbɑB
	array_insert(ar, br);
	array_insert(ai, bi);
	fft2(br, bi, N, N, flag);
}

void function(double br[], double bi[], double fr[], double fi[], double t){//̉EӂvZBfbkƂƂ̌vZʁB
	double cr[N*N], ci[N*N];
	int i, k1, k2;
	do_fft2(br, bi, cr, ci, 1);
	for(i=0; i<N*N; ++i){
		cr[i] = cr[i]*cr[i]*cr[i] - 3*cr[i]*ci[i]*ci[i];
		ci[i] = 3*cr[i]*cr[i]*ci[i] - ci[i]*ci[i]*ci[i];
	}
	do_fft2(cr, ci, fr, fi, 0);
	for(i=0; i<N*N; ++i){
		k1 = i/N;
		k2 = i%N;
		fr[i] = -alpha*fr[i] + (alpha - beta*(k1*k1+k2*k2) - gamma*(4-2*M_PI*sqrt(k1*k1+k2*k2)))*br[i];
		if(i==0){fr[i] = fr[i] + hini - v*t;}
		fi[i] = -alpha*fi[i] + (alpha - beta*(k1*k1+k2*k2) - gamma*(4-2*M_PI*sqrt(k1*k1+k2*k2)))*bi[i];
	}
}

void runge2(double br[], double bi[], double t){//2̃QNb^@̂
	double q1r[N*N], q1i[N*N], q2r[N*N], q2i[N*N], b2r[N*N], b2i[N*N];
	int i;
	function(br, bi, q1r, q1i, t);
	for(i=0; i<N*N; ++i){
		b2r[i] = br[i] + q1r[i]*dt;
		b2i[i] = bi[i] + q1i[i]*dt;
	}
	function(b2r, b2i, q2r, q2i, t+dt);
	for(i=0; i<N*N; ++i){
		br[i] = br[i] + (q1r[i]+q2r[i])*dt/2;
		bi[i] = bi[i] + (q1i[i]+q2i[i])*dt/2;
	}
}


